/*
  Copyright (c) 2022-2023, Intel Corporation

  SPDX-License-Identifier: BSD-3-Clause
*/

/** @file version.h
    @brief defines the ISPC version
*/

#pragma once

#define ISPC_VERSION_MAJOR 1
#define ISPC_VERSION_MINOR 22
#define ISPC_VERSION_PATCH 0
#define ISPC_VERSION "1.22.0"
