// Copyright 2022-2023 Intel Corporation

#include "version.h"

1 VERSIONINFO
 FILEVERSION ISPC_VERSION_MAJOR,ISPC_VERSION_MINOR,ISPC_VERSION_PATCH,0
 PRODUCTVERSION ISPC_VERSION_MAJOR,ISPC_VERSION_MINOR,ISPC_VERSION_PATCH,0
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "Intel Corporation"
            VALUE "FileDescription", "Intel(R) Implicit SPMD Program Compiler"
            VALUE "FileVersion", ISPC_VERSION
            VALUE "ProductVersion", ISPC_VERSION
            VALUE "LegalCopyright", "Copyright 2010-2023 Intel Corporation"
            VALUE "InternalName", "Intel(R) ISPC"
            VALUE "ProductName", "Intel(R) ISPC"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
